@@ -32,11 +32,12 @@ class Schema;
 class StructSchema;
 class EnumSchema;
 class InterfaceSchema;
+class ConstSchema;
 class ListSchema;
 
 template <typename T, Kind k = kind<T>()> struct SchemaType_ { typedef Schema Type; };
-template <typename T> struct SchemaType_<T, Kind::PRIMITIVE> { typedef schema::Type::Body::Which Type; };
-template <typename T> struct SchemaType_<T, Kind::BLOB> { typedef schema::Type::Body::Which Type; };
+template <typename T> struct SchemaType_<T, Kind::PRIMITIVE> { typedef schema::Type::Which Type; };
+template <typename T> struct SchemaType_<T, Kind::BLOB> { typedef schema::Type::Which Type; };
 template <typename T> struct SchemaType_<T, Kind::ENUM> { typedef EnumSchema Type; };
 template <typename T> struct SchemaType_<T, Kind::STRUCT> { typedef StructSchema Type; };
 template <typename T> struct SchemaType_<T, Kind::INTERFACE> { typedef InterfaceSchema Type; };
@@ -46,17 +47,29 @@ template <typename T>
 using SchemaType = typename SchemaType_<T>::Type;
 // SchemaType<T> is the type of T's schema, e.g. StructSchema if T is a struct.
 
+namespace _ {  // private
+extern const RawSchema NULL_SCHEMA;
+extern const RawSchema NULL_STRUCT_SCHEMA;
+extern const RawSchema NULL_ENUM_SCHEMA;
+extern const RawSchema NULL_INTERFACE_SCHEMA;
+extern const RawSchema NULL_CONST_SCHEMA;
+// The schema types default to these null (empty) schemas in case of error, especially when
+// exceptions are disabled.
+}  // namespace _ (private)
+
 class Schema {
   // Convenience wrapper around capnp::schema::Node.
 
 public:
-  inline Schema(): raw(nullptr) {}
+  inline Schema(): raw(&_::NULL_SCHEMA) {}
 
   template <typename T>
   static inline SchemaType<T> from() { return SchemaType<T>::template fromImpl<T>(); }
   // Get the Schema for a particular compiled-in type.
 
   schema::Node::Reader getProto() const;
+  // Get the underlying Cap'n Proto representation of the schema node.  (Note that this accessor
+  // has performance comparable to accessors of struct-typed fields on Reader classes.)
 
   kj::ArrayPtr<const word> asUncheckedMessage() const;
   // Get the encoded schema node content as a single message segment.  It is safe to read as an
@@ -65,13 +78,28 @@ public:
   Schema getDependency(uint64_t id) const;
   // Gets the Schema for one of this Schema's dependencies.  For example, if this Schema is for a
   // struct, you could look up the schema for one of its fields' types.  Throws an exception if this
-  // schema doesn't actually depend on the given id.  Note that annotation declarations are not
-  // considered dependencies for this purpose.
+  // schema doesn't actually depend on the given id.
+  //
+  // Note that not all type IDs found in the schema node are considered "dependencies" -- only the
+  // ones that are needed to implement the dynamic API are.  That includes:
+  // - Field types.
+  // - Group types.
+  // - scopeId for group nodes, but NOT otherwise.
+  // - Method parameter and return types.
+  //
+  // The following are NOT considered dependencies:
+  // - Nested nodes.
+  // - scopeId for a non-group node.
+  // - Annotations.
+  //
+  // To obtain schemas for those, you would need a SchemaLoader.
 
   StructSchema asStruct() const;
   EnumSchema asEnum() const;
   InterfaceSchema asInterface() const;
-  // Cast the Schema to a specific type.  Throws an exception if the type doesn't match.
+  ConstSchema asConst() const;
+  // Cast the Schema to a specific type.  Throws an exception if the type doesn't match.  Use
+  // getProto() to determine type, e.g. getProto().isStruct().
 
   inline bool operator==(const Schema& other) const { return raw == other.raw; }
   inline bool operator!=(const Schema& other) const { return raw != other.raw; }
@@ -87,6 +115,9 @@ public:
   // - This schema was loaded with SchemaLoader, the type ID matches typeId<T>(), and
   //   loadCompiledTypeAndDependencies<T>() was called on the SchemaLoader.
 
+  kj::StringPtr getShortDisplayName() const;
+  // Get the short version of the node's display name.
+
 private:
   const _::RawSchema* raw;
 
@@ -101,9 +132,12 @@ private:
 
   void requireUsableAs(const _::RawSchema* expected) const;
 
+  uint32_t getSchemaOffset(const schema::Value::Reader& value) const;
+
   friend class StructSchema;
   friend class EnumSchema;
   friend class InterfaceSchema;
+  friend class ConstSchema;
   friend class ListSchema;
   friend class SchemaLoader;
 };
@@ -112,20 +146,39 @@ private:
 
 class StructSchema: public Schema {
 public:
-  StructSchema() = default;
+  inline StructSchema(): Schema(&_::NULL_STRUCT_SCHEMA) {}
 
-  class Member;
   class Field;
-  class Union;
-  class Group;
-  class MemberList;
+  class FieldList;
+  class FieldSubset;
 
-  MemberList getMembers() const;
+  FieldList getFields() const;
+  // List top-level fields of this struct.  This list will contain top-level groups (including
+  // named unions) but not the members of those groups.  The list does, however, contain the
+  // members of the unnamed union, if there is one.
 
-  kj::Maybe<Member> findMemberByName(kj::StringPtr name) const;
+  FieldSubset getUnionFields() const;
+  // If the field contains an unnamed union, get a list of fields in the union, ordered by
+  // ordinal.  Since discriminant values are assigned sequentially by ordinal, you may index this
+  // list by discriminant value.
 
-  Member getMemberByName(kj::StringPtr name) const;
-  // Like findMemberByName() but throws an exception on failure.
+  FieldSubset getNonUnionFields() const;
+  // Get the fields of this struct which are not in an unnamed union, ordered by ordinal.
+
+  kj::Maybe<Field> findFieldByName(kj::StringPtr name) const;
+  // Find the field with the given name, or return null if there is no such field.  If the struct
+  // contains an unnamed union, then this will find fields of that union in addition to fields
+  // of the outer struct, since they exist in the same namespace.  It will not, however, find
+  // members of groups (including named unions) -- you must first look up the group itself,
+  // then dig into its type.
+
+  Field getFieldByName(kj::StringPtr name) const;
+  // Like findFieldByName() but throws an exception on failure.
+
+  kj::Maybe<Field> getFieldByDiscriminant(uint16_t discriminant) const;
+  // Finds the field whose `discriminantValue` is equal to the given value, or returns null if
+  // there is no such field.  (If the schema does not represent a union or a struct containing
+  // an unnamed union, then this always returns null.)
 
 private:
   StructSchema(const _::RawSchema* raw): Schema(raw) {}
@@ -135,51 +188,17 @@ private:
   friend class Schema;
   friend kj::StringTree _::structString(
       _::StructReader reader, const _::RawSchema& schema);
-  friend kj::StringTree _::unionString(
-      _::StructReader reader, const _::RawSchema& schema, uint memberIndex);
 };
 
-class StructSchema::Member {
+class StructSchema::Field {
 public:
-  Member() = default;
+  Field() = default;
 
-  inline schema::StructNode::Member::Reader getProto() const { return proto; }
+  inline schema::Field::Reader getProto() const { return proto; }
   inline StructSchema getContainingStruct() const { return parent; }
 
   inline uint getIndex() const { return index; }
-  // Get the index of this member within the containing struct or union.
-
-  kj::Maybe<Union> getContainingUnion() const;
-  // If this a member of a union, gets the containing union schema.
-
-  Field asField() const;
-  // Cast the member to a Field.  Throws an exception if not a field.
-
-  Union asUnion() const;
-  // Cast the member to a Union.  Throws an exception if not a union.
-
-  Group asGroup() const;
-  // Cast the member to a Group.  Throws an exception if not a group.
-
-  inline bool operator==(const Member& other) const;
-  inline bool operator!=(const Member& other) const { return !(*this == other); }
-
-private:
-  StructSchema parent;
-  uint unionIndex;  // 0 = none, >0 = actual union index - 1
-  uint index;
-  schema::StructNode::Member::Reader proto;
-
-  inline Member(StructSchema parent, uint unionIndex, uint index,
-                schema::StructNode::Member::Reader proto)
-      : parent(parent), unionIndex(unionIndex), index(index), proto(proto) {}
-
-  friend class StructSchema;
-};
-
-class StructSchema::Field: public Member {
-public:
-  Field() = default;
+  // Get the index of this field within the containing struct or union.
 
   uint32_t getDefaultValueSchemaOffset() const;
   // For struct, list, and object fields, returns the offset, in words, within the first segment of
@@ -201,63 +220,63 @@ public:
   //
   // If the above does not make sense, you probably don't need this method.
 
+  inline bool operator==(const Field& other) const;
+  inline bool operator!=(const Field& other) const { return !(*this == other); }
+
 private:
-  inline Field(const Member& base): Member(base) {}
+  StructSchema parent;
+  uint index;
+  schema::Field::Reader proto;
+
+  inline Field(StructSchema parent, uint index, schema::Field::Reader proto)
+      : parent(parent), index(index), proto(proto) {}
 
   friend class StructSchema;
 };
 
-class StructSchema::Union: public Member {
+class StructSchema::FieldList {
 public:
-  Union() = default;
+  FieldList() = default;  // empty list
 
-  MemberList getMembers() const;
-
-  kj::Maybe<Member> findMemberByName(kj::StringPtr name) const;
-
-  Member getMemberByName(kj::StringPtr name) const;
-  // Like findMemberByName() but throws an exception on failure.
-
-private:
-  inline Union(const Member& base): Member(base) {}
-
-  friend class StructSchema;
-};
-
-class StructSchema::Group: public Member {
-public:
-  Group() = default;
-
-  MemberList getMembers() const;
-
-  kj::Maybe<Member> findMemberByName(kj::StringPtr name) const;
-
-  Member getMemberByName(kj::StringPtr name) const;
-  // Like findMemberByName() but throws an exception on failure.
-
-private:
-  inline Group(const Member& base): Member(base) {}
-
-  friend class StructSchema;
-};
-
-class StructSchema::MemberList {
-public:
   inline uint size() const { return list.size(); }
-  inline Member operator[](uint index) const { return Member(parent, unionIndex, index, list[index]); }
+  inline Field operator[](uint index) const { return Field(parent, index, list[index]); }
 
-  typedef _::IndexingIterator<const MemberList, Member> Iterator;
+  typedef _::IndexingIterator<const FieldList, Field> Iterator;
   inline Iterator begin() const { return Iterator(this, 0); }
   inline Iterator end() const { return Iterator(this, size()); }
 
 private:
   StructSchema parent;
-  uint unionIndex;
-  List<schema::StructNode::Member>::Reader list;
+  List<schema::Field>::Reader list;
 
-  inline MemberList(StructSchema parent, uint unionIndex,
-                    List<schema::StructNode::Member>::Reader list)
-      : parent(parent), unionIndex(unionIndex), list(list) {}
+  inline FieldList(StructSchema parent, List<schema::Field>::Reader list)
+      : parent(parent), list(list) {}
+
+  friend class StructSchema;
+};
+
+class StructSchema::FieldSubset {
+public:
+  FieldSubset() = default;  // empty list
+
+  inline uint size() const { return size_; }
+  inline Field operator[](uint index) const {
+    return Field(parent, indices[index], list[indices[index]]);
+  }
+
+  typedef _::IndexingIterator<const FieldSubset, Field> Iterator;
+  inline Iterator begin() const { return Iterator(this, 0); }
+  inline Iterator end() const { return Iterator(this, size()); }
+
+private:
+  StructSchema parent;
+  List<schema::Field>::Reader list;
+  const uint16_t* indices;
+  uint size_;
+
+  inline FieldSubset(StructSchema parent, List<schema::Field>::Reader list,
+                     const uint16_t* indices, uint size)
+      : parent(parent), list(list), indices(indices), size_(size) {}
 
   friend class StructSchema;
 };
@@ -266,7 +285,7 @@ private:
 
 class EnumSchema: public Schema {
 public:
-  EnumSchema() = default;
+  inline EnumSchema(): Schema(&_::NULL_ENUM_SCHEMA) {}
 
   class Enumerant;
   class EnumerantList;
@@ -290,7 +309,7 @@ class EnumSchema::Enumerant {
 public:
   Enumerant() = default;
 
-  inline schema::EnumNode::Enumerant::Reader getProto() const { return proto; }
+  inline schema::Enumerant::Reader getProto() const { return proto; }
   inline EnumSchema getContainingEnum() const { return parent; }
 
   inline uint16_t getOrdinal() const { return ordinal; }
@@ -302,9 +321,9 @@ public:
 private:
   EnumSchema parent;
   uint16_t ordinal;
-  schema::EnumNode::Enumerant::Reader proto;
+  schema::Enumerant::Reader proto;
 
-  inline Enumerant(EnumSchema parent, uint16_t ordinal, schema::EnumNode::Enumerant::Reader proto)
+  inline Enumerant(EnumSchema parent, uint16_t ordinal, schema::Enumerant::Reader proto)
       : parent(parent), ordinal(ordinal), proto(proto) {}
 
   friend class EnumSchema;
@@ -312,6 +331,8 @@ private:
 
 class EnumSchema::EnumerantList {
 public:
+  EnumerantList() = default;  // empty list
+
   inline uint size() const { return list.size(); }
   inline Enumerant operator[](uint index) const { return Enumerant(parent, index, list[index]); }
 
@@ -321,9 +342,9 @@ public:
 
 private:
   EnumSchema parent;
-  List<schema::EnumNode::Enumerant>::Reader list;
+  List<schema::Enumerant>::Reader list;
 
-  inline EnumerantList(EnumSchema parent, List<schema::EnumNode::Enumerant>::Reader list)
+  inline EnumerantList(EnumSchema parent, List<schema::Enumerant>::Reader list)
       : parent(parent), list(list) {}
 
   friend class EnumSchema;
@@ -333,7 +354,7 @@ private:
 
 class InterfaceSchema: public Schema {
 public:
-  InterfaceSchema() = default;
+  inline InterfaceSchema(): Schema(&_::NULL_INTERFACE_SCHEMA) {}
 
   class Method;
   class MethodList;
@@ -345,19 +366,32 @@ public:
   Method getMethodByName(kj::StringPtr name) const;
   // Like findMethodByName() but throws an exception on failure.
 
+  bool extends(InterfaceSchema other) const;
+  // Returns true if `other` is a superclass of this interface (including if `other == *this`).
+
+  kj::Maybe<InterfaceSchema> findSuperclass(uint64_t typeId) const;
+  // Find the superclass of this interface with the given type ID.  Returns null if the interface
+  // extends no such type.
+
 private:
   InterfaceSchema(const _::RawSchema* raw): Schema(raw) {}
   template <typename T> static inline InterfaceSchema fromImpl() {
     return InterfaceSchema(&_::rawSchema<T>());
   }
   friend class Schema;
+
+  kj::Maybe<Method> findMethodByName(kj::StringPtr name, uint& counter) const;
+  bool extends(InterfaceSchema other, uint& counter) const;
+  kj::Maybe<InterfaceSchema> findSuperclass(uint64_t typeId, uint& counter) const;
+  // We protect against malicious schemas with large or cyclic hierarchies by cutting off the
+  // search when the counter reaches a threshold.
 };
 
 class InterfaceSchema::Method {
 public:
   Method() = default;
 
-  inline schema::InterfaceNode::Method::Reader getProto() const { return proto; }
+  inline schema::Method::Reader getProto() const { return proto; }
   inline InterfaceSchema getContainingInterface() const { return parent; }
 
   inline uint16_t getOrdinal() const { return ordinal; }
@@ -369,10 +403,10 @@ public:
 private:
   InterfaceSchema parent;
   uint16_t ordinal;
-  schema::InterfaceNode::Method::Reader proto;
+  schema::Method::Reader proto;
 
   inline Method(InterfaceSchema parent, uint16_t ordinal,
-                schema::InterfaceNode::Method::Reader proto)
+                schema::Method::Reader proto)
       : parent(parent), ordinal(ordinal), proto(proto) {}
 
   friend class InterfaceSchema;
@@ -380,6 +414,8 @@ private:
 
 class InterfaceSchema::MethodList {
 public:
+  MethodList() = default;  // empty list
+
   inline uint size() const { return list.size(); }
   inline Method operator[](uint index) const { return Method(parent, index, list[index]); }
 
@@ -389,9 +425,9 @@ public:
 
 private:
   InterfaceSchema parent;
-  List<schema::InterfaceNode::Method>::Reader list;
+  List<schema::Method>::Reader list;
 
-  inline MethodList(InterfaceSchema parent, List<schema::InterfaceNode::Method>::Reader list)
+  inline MethodList(InterfaceSchema parent, List<schema::Method>::Reader list)
       : parent(parent), list(list) {}
 
   friend class InterfaceSchema;
@@ -399,6 +435,32 @@ private:
 
 // -------------------------------------------------------------------
 
+class ConstSchema: public Schema {
+  // Represents a constant declaration.
+  //
+  // `ConstSchema` can be implicitly cast to DynamicValue to read its value.
+
+public:
+  inline ConstSchema(): Schema(&_::NULL_CONST_SCHEMA) {}
+
+  template <typename T>
+  ReaderFor<T> as() const;
+  // Read the constant's value.  This is a convenience method equivalent to casting the ConstSchema
+  // to a DynamicValue and then calling its `as<T>()` method.  For dependency reasons, this method
+  // is defined in <capnp/dynamic.h>, which you must #include explicitly.
+
+  uint32_t getValueSchemaOffset() const;
+  // Much like StructSchema::Field::getDefaultValueSchemaOffset(), if the constant has pointer
+  // type, this gets the offset from the beginning of the constant's schema node to a pointer
+  // representing the constant value.
+
+private:
+  ConstSchema(const _::RawSchema* raw): Schema(raw) {}
+  friend class Schema;
+};
+
+// -------------------------------------------------------------------
+
 class ListSchema {
   // ListSchema is a little different because list types are not described by schema nodes.  So,
   // ListSchema doesn't subclass Schema.
@@ -406,7 +468,7 @@ class ListSchema {
 public:
   ListSchema() = default;
 
-  static ListSchema of(schema::Type::Body::Which primitiveType);
+  static ListSchema of(schema::Type::Which primitiveType);
   static ListSchema of(StructSchema elementType);
   static ListSchema of(EnumSchema elementType);
   static ListSchema of(InterfaceSchema elementType);
@@ -417,7 +479,7 @@ public:
   // Construct from an element type schema.  Requires a context which can handle getDependency()
   // requests for any type ID found in the schema.
 
-  inline schema::Type::Body::Which whichElementType() const;
+  inline schema::Type::Which whichElementType() const;
   // Get the element type's "which()".  ListSchema does not actually store a schema::Type::Reader
   // describing the element type, but if it did, this would be equivalent to calling
   // .getBody().which() on that type.
@@ -436,15 +498,15 @@ public:
   void requireUsableAs() const;
 
 private:
-  schema::Type::Body::Which elementType;
+  schema::Type::Which elementType;
   uint8_t nestingDepth;  // 0 for T, 1 for List(T), 2 for List(List(T)), ...
   Schema elementSchema;  // if elementType is struct, enum, interface...
 
-  inline ListSchema(schema::Type::Body::Which elementType)
+  inline ListSchema(schema::Type::Which elementType)
       : elementType(elementType), nestingDepth(0) {}
-  inline ListSchema(schema::Type::Body::Which elementType, Schema elementSchema)
+  inline ListSchema(schema::Type::Which elementType, Schema elementSchema)
       : elementType(elementType), nestingDepth(0), elementSchema(elementSchema) {}
-  inline ListSchema(schema::Type::Body::Which elementType, uint8_t nestingDepth,
+  inline ListSchema(schema::Type::Which elementType, uint8_t nestingDepth,
                     Schema elementSchema)
       : elementType(elementType), nestingDepth(nestingDepth), elementSchema(elementSchema) {}
 
@@ -462,28 +524,28 @@ private:
 // =======================================================================================
 // inline implementation
 
-template <> inline schema::Type::Body::Which Schema::from<Void>() { return schema::Type::Body::VOID_TYPE; }
-template <> inline schema::Type::Body::Which Schema::from<bool>() { return schema::Type::Body::BOOL_TYPE; }
-template <> inline schema::Type::Body::Which Schema::from<int8_t>() { return schema::Type::Body::INT8_TYPE; }
-template <> inline schema::Type::Body::Which Schema::from<int16_t>() { return schema::Type::Body::INT16_TYPE; }
-template <> inline schema::Type::Body::Which Schema::from<int32_t>() { return schema::Type::Body::INT32_TYPE; }
-template <> inline schema::Type::Body::Which Schema::from<int64_t>() { return schema::Type::Body::INT64_TYPE; }
-template <> inline schema::Type::Body::Which Schema::from<uint8_t>() { return schema::Type::Body::UINT8_TYPE; }
-template <> inline schema::Type::Body::Which Schema::from<uint16_t>() { return schema::Type::Body::UINT16_TYPE; }
-template <> inline schema::Type::Body::Which Schema::from<uint32_t>() { return schema::Type::Body::UINT32_TYPE; }
-template <> inline schema::Type::Body::Which Schema::from<uint64_t>() { return schema::Type::Body::UINT64_TYPE; }
-template <> inline schema::Type::Body::Which Schema::from<float>() { return schema::Type::Body::FLOAT32_TYPE; }
-template <> inline schema::Type::Body::Which Schema::from<double>() { return schema::Type::Body::FLOAT64_TYPE; }
-template <> inline schema::Type::Body::Which Schema::from<Text>() { return schema::Type::Body::TEXT_TYPE; }
-template <> inline schema::Type::Body::Which Schema::from<Data>() { return schema::Type::Body::DATA_TYPE; }
+template <> inline schema::Type::Which Schema::from<Void>() { return schema::Type::VOID; }
+template <> inline schema::Type::Which Schema::from<bool>() { return schema::Type::BOOL; }
+template <> inline schema::Type::Which Schema::from<int8_t>() { return schema::Type::INT8; }
+template <> inline schema::Type::Which Schema::from<int16_t>() { return schema::Type::INT16; }
+template <> inline schema::Type::Which Schema::from<int32_t>() { return schema::Type::INT32; }
+template <> inline schema::Type::Which Schema::from<int64_t>() { return schema::Type::INT64; }
+template <> inline schema::Type::Which Schema::from<uint8_t>() { return schema::Type::UINT8; }
+template <> inline schema::Type::Which Schema::from<uint16_t>() { return schema::Type::UINT16; }
+template <> inline schema::Type::Which Schema::from<uint32_t>() { return schema::Type::UINT32; }
+template <> inline schema::Type::Which Schema::from<uint64_t>() { return schema::Type::UINT64; }
+template <> inline schema::Type::Which Schema::from<float>() { return schema::Type::FLOAT32; }
+template <> inline schema::Type::Which Schema::from<double>() { return schema::Type::FLOAT64; }
+template <> inline schema::Type::Which Schema::from<Text>() { return schema::Type::TEXT; }
+template <> inline schema::Type::Which Schema::from<Data>() { return schema::Type::DATA; }
 
 template <typename T>
 inline void Schema::requireUsableAs() const {
   requireUsableAs(&_::rawSchema<T>());
 }
 
-inline bool StructSchema::Member::operator==(const Member& other) const {
-  return parent == other.parent && unionIndex == other.unionIndex && index == other.index;
+inline bool StructSchema::Field::operator==(const Field& other) const {
+  return parent == other.parent && index == other.index;
 }
 inline bool EnumSchema::Enumerant::operator==(const Enumerant& other) const {
   return parent == other.parent && ordinal == other.ordinal;
@@ -493,21 +555,21 @@ inline bool InterfaceSchema::Method::operator==(const Method& other) const {
 }
 
 inline ListSchema ListSchema::of(StructSchema elementType) {
-  return ListSchema(schema::Type::Body::STRUCT_TYPE, 0, elementType);
+  return ListSchema(schema::Type::STRUCT, 0, elementType);
 }
 inline ListSchema ListSchema::of(EnumSchema elementType) {
-  return ListSchema(schema::Type::Body::ENUM_TYPE, 0, elementType);
+  return ListSchema(schema::Type::ENUM, 0, elementType);
 }
 inline ListSchema ListSchema::of(InterfaceSchema elementType) {
-  return ListSchema(schema::Type::Body::INTERFACE_TYPE, 0, elementType);
+  return ListSchema(schema::Type::INTERFACE, 0, elementType);
 }
 inline ListSchema ListSchema::of(ListSchema elementType) {
   return ListSchema(elementType.elementType, elementType.nestingDepth + 1,
                     elementType.elementSchema);
 }
 
-inline schema::Type::Body::Which ListSchema::whichElementType() const {
-  return nestingDepth == 0 ? elementType : schema::Type::Body::LIST_TYPE;
+inline schema::Type::Which ListSchema::whichElementType() const {
+  return nestingDepth == 0 ? elementType : schema::Type::LIST;
 }
 
 inline bool ListSchema::operator==(const ListSchema& other) const {
